from collections.abc import Callable
import olefile
import os
import shutil
from typing import Optional
import xml.etree.ElementTree as ET

from . import decompress
from . import types
from . import util

MER_FTLINX_ATTRIB_FILTER = {'name', 'address', 'portNumber', 'NATPrivateAddress'}

def apa_to_med(
    input_path: str | bytes,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[types.MEArchive]:
    # Application-specific handling for *.APA files that
    # keeps streams in memory.
    #
    # Just a placeholder function for exploration for now
    with olefile.OleFileIO(input_path) as ole:
        streams = decompress.decompress_archive(
            ole=ole,
            progress=progress
        )
        return streams
    
def apa_to_med_folder(
    input_path: str | bytes, 
    output_path: str,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    # Application-specific handling for *.APA files that
    # writes streams to a folder.
    #
    # Just a placeholder function for exploration for now
    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)
    streams = apa_to_med(
        input_path=input_path,
        progress=progress
    )
    for stream in streams:
        stream_output_path = decompress._create_subfolders(output_path, stream.path)
        with open(stream_output_path, 'wb') as f:
            f.write(stream.data)

def mer_to_med(
    input_path: str | bytes,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[types.MEArchive]:
    # Application-specific handling for *.MER files that
    # keeps streams in memory.
    #
    # Just a placeholder function for exploration for now
    with olefile.OleFileIO(input_path) as ole:
        streams = decompress.decompress_archive(
            ole=ole,
            progress=progress
        )
        return streams
    
def mer_to_med_folder(
    input_path: str | bytes,
    output_path: str,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    # Application-specific handling for *.MER files that
    # writes streams to a folder.
    #
    # Just a placeholder function for exploration for now
    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)
    streams = mer_to_med(
        input_path=input_path,
        progress=progress
    )
    for stream in streams:
        stream_output_path = decompress._create_subfolders(output_path, stream.path)
        with open(stream_output_path, 'wb') as f:
            f.write(stream.data)

def _mer_get_shortcut_names(
    streams: list[types.MEArchive],
) -> list[tuple[str, str]]:
    file = 'RSLinx Enterprise/SCLocal.xml'
    try:
        xml = util._get_stream_by_name(streams, file).data.decode('utf-16-le')
    except Exception as e:
        raise FileNotFoundError(f'{file} was not found.  Shortcuts may not exist.')

    root = ET.fromstring(xml)
    shortcuts = []
    for shortcut in root.findall('shortcut'):
        name = shortcut.attrib.get('name')
        device = shortcut.attrib.get('device')
        shortcuts.append((name, device))
    return shortcuts

def _mer_filter_shortcut_attributes(element: ET.Element, whitelist: list[str]) -> ET.Element:
    # Create a new Element with the same tag
    filtered_node = ET.Element(element.tag)

    # Copy only specific attributes
    for attr in whitelist:
        if attr in element.attrib:
            filtered_node.set(attr, element.attrib[attr])

    return filtered_node

def _mer_get_device_nodes(topology: ET.Element, node1: str, node2: str) -> list[ET.Element]:
    parent_map = {child: parent for parent in topology.iter() for child in parent}
    for device in topology.findall(".//device[@name='" + node2 + "']"):
        current = device
        path_nodes = []
        found_node1 = False
        while current is not None:
            name = current.attrib.get("name")
            if name:
                path_nodes.insert(0, current)
                if name == node1:
                    found_node1 = True
            current = parent_map.get(current)
        if found_node1:
            return path_nodes
    return []

def _mer_get_shortcut_nodes(
    streams: list[types.MEArchive],
    shortcuts: list[tuple[str, str]]
) -> list[tuple[str, list[ET.Element]]]:
    file = 'RSLinx Enterprise/RSLinxNG.xml'
    try:
        xml = util._get_stream_by_name(streams, 'RSLinx Enterprise/RSLinxNG.xml').data.decode('utf-16-le')
    except Exception as e:
        raise FileNotFoundError(f'{file} was not found.  Shortcuts may not exist.')

    root = ET.fromstring(xml)
    topology = root.find("Topology")
    result = []
    for shortcut_name, device_string in shortcuts:
        if '.' not in device_string:
            result.append((shortcut_name, ""))
            continue
        node1, node2 = device_string.split('.', 1)
        path = _mer_get_device_nodes(topology, node1, node2)
        result.append((shortcut_name, path))

    return result

def _mer_print_shortcut_nodes(
    shortcut: str,
    nodes: list[ET.Element]
):
    print(f'Shortcut {shortcut} Topology:')
    for i, line in enumerate(nodes):
        # Determine prefix based on position
        if i == len(nodes) - 1:
            connector = "└── "
        else:
            connector = "├── "
        
        # Build indentation
        indent = "│   " * i
        print(f"{indent}{connector}{line}")

def mer_get_shortcuts(
    input_path: str | bytes,
    print_summary: bool = False,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[tuple[str, list[ET.Element]]]:
    
    # Application-specific function for *.MER files to
    # get the list of communications shortcuts    
    streams = mer_to_med(
        input_path=input_path,
        progress=progress
    )
    shortcuts = _mer_get_shortcut_names(streams)
    paths = _mer_get_shortcut_nodes(streams, shortcuts)
    for (name, path) in paths:
        filtered_nodes = []
        for node in path:
            filtered_node =_mer_filter_shortcut_attributes(node, MER_FTLINX_ATTRIB_FILTER)
            filtered_nodes.append(ET.tostring(filtered_node, encoding='unicode'))
        if print_summary: _mer_print_shortcut_nodes(name, filtered_nodes)

    return paths

def _mer_rewrite_checksum(
    input_path: str,
    output_path: str
) -> bytes:
    # Read file and calculate checksum
    file_data = bytes()
    with open(input_path, 'rb') as f:
        raw_data = f.read()
        file_data = raw_data[:-4]
    new_checksum = util.crc32_checksum(file_data).to_bytes(length=4, byteorder='little', signed=False)
    file_data = file_data + new_checksum

    # Re-write file with updated checksum
    with open(output_path, 'wb') as f:
        f.write(file_data)

    return new_checksum

def mer_unlock(
    input_path: str,
    output_path: str,
):
    # Create copy of *.MER file to work with
    if not(os.path.exists(os.path.dirname(output_path))): os.makedirs(os.path.dirname(output_path), exist_ok=True)
    shutil.copy(input_path, output_path)

    # Edit the contents of FILE_PROTECTION
    with olefile.OleFileIO(output_path, write_mode=True) as ole:
        stream = bytearray(ole.openstream('FILE_PROTECTION').read())
        msb = len(stream)

        if (msb < 7): raise FileNotFoundError(f'Unexpected file size {msb} for FILE_PROTECTION, cannot proceed.')
        stream[0] = 0x00 # No Password
        stream[1] = 0x03 # Length LSW?
        stream[2] = 0x00 # Length MSW?
        stream[3] = 0x00 # Allow Convert
        for i in range(4,msb):
            stream[i] = 0x00 # Remaining content would be password.

        ole.write_stream('FILE_PROTECTION', bytes(stream))

    # Re-rewrite the checksum of the *.MER copy in place
    _mer_rewrite_checksum(input_path=output_path, output_path=output_path)

def apa_unlock(
    input_path: str,
    output_path: str
):
    # Create copy of *.APA file to work with
    if not(os.path.exists(os.path.dirname(output_path))): os.makedirs(os.path.dirname(output_path), exist_ok=True)
    shutil.copy(input_path, output_path)

    # Edit the contents of FTSPHasPwd
    with olefile.OleFileIO(output_path, write_mode=True) as ole:
        stream = bytearray(ole.openstream('FTSPHasPwd').read())
        msb = len(stream)

        if (msb < 4): raise FileNotFoundError(f'Unexpected file size {msb} for FTSPHasPwd, cannot proceed.')
        stream[0] = 0x00 # No Password
        ole.write_stream('FTSPHasPwd', bytes(stream))



from enum import StrEnum

from .. import comms
from . import helper
from . import types

# Known static value for successful creation of a folder.
# Further investigation needed.
CREATE_DIR_SUCCESS = 0

# Known functions available from FuwHelper.
class FuwHelperFunctions(StrEnum):
    CLEAR_FOLDER = 'ClearRemDirectory', # Args: {Folder Path}
    CREATE_FOLDER = 'CreateRemDirectory' # Args: {Folder Path}, Returns: Static value if successful [Further investigation needed]
    DELETE_FILE = 'DeleteRemFile' # Args: {File Path}, Returns: ???
    DELETE_FOLDER = 'DeleteRemDirectory' # Args: {Folder Path}
    DISABLE_ME_CORRUPT_SCREEN = 'DisableMECorruptScreen' # Args: Null
    DISABLE_SCREENSAVER = 'DisableScreenSaver' # Args: Null
    ENABLE_ME_CORRUPT_SCREEN = 'EnableMECorruptScreen' # Args: Null
    ENABLE_SCREENSAVER = 'EnableScreenSaver' # Args: Null
    GET_FILE_EXISTS = 'FileExists' # Args: {File Path}, Returns: {File Size} if {File Path} exists, 0 otherwise
    GET_FOLDER_EXISTS = 'StorageExists' # Args: {Folder Path}, Returns: 1 if {Folder Path} exists
    GET_FREE_SPACE = 'FreeSpace' # Args: {Folder Path}, Returns: Free size in bytes
    GET_PROCESS_RUNNING = 'IsExeRunning' # Args: {Process Name}, Resp Code: 0 successful, Resp Data: 1 if process is running, 0 otherwise
    GET_TERMINAL_PARTITION_SIZE = 'GetTerminalPartitionSize' # Args: Null, Retruns {Partition Size} in bytes
    GET_TERMINAL_OS_REV = 'GetTerminalOSRev' # Args: Null, Returns: Build String
    GET_TOTAL_SPACE = 'TotalSpace' # Args: {Folder Path}, Returns: Free size in bytes
    START_PROCESS = 'InvokeEXE' # Args: {Binary Path}
    STOP_PROCESS = 'TerminateEXE' # Args: {Binary Name}
    STOP_PROCESS_ME = 'SafeTerminateME' # Args: Null, Resp Code: 0 successful, 2 failed/wasn't running, Resp Data: Null

def clear_folder(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.CLEAR_FOLDER, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return True

def create_folder(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.CREATE_FOLDER, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != CREATE_DIR_SUCCESS): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_file(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.DELETE_FILE, file_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete file on terminal: {file_path}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_folder(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.DELETE_FOLDER, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete folder on terminal: {folder_path}, response code: {resp_code}, response data: {resp_data}.')
    return True

def get_process_running(cip: comms.Driver, paths: types.MEPaths, process_name: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_PROCESS_RUNNING, process_name]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_file_exists(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_FILE_EXISTS, file_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_folder_exists(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_FOLDER_EXISTS, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_free_space(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> int:
    if not(get_folder_exists(cip, paths, folder_path)): raise FileNotFoundError(f'Folder {folder_path} does not exist on remote terminal.')
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_FREE_SPACE, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return int(resp_data)

def get_os_rev(cip: comms.Driver, paths: types.MEPaths) -> str:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_TERMINAL_OS_REV, '']
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return str(resp_data)

def get_partition_size(cip: comms.Driver, paths: types.MEPaths) -> int:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_TERMINAL_PARTITION_SIZE, '']
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return int(resp_data)

def get_total_space(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> int:
    if not(get_folder_exists(cip, paths, folder_path)): raise FileNotFoundError(f'Folder {folder_path} does not exist on remote terminal.')
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_TOTAL_SPACE, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return int(resp_data)

def set_me_corrupt_screen(cip: comms.Driver, paths: types.MEPaths, state: bool) -> bool:
    if state:
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.ENABLE_ME_CORRUPT_SCREEN, '']
    else:
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.DISABLE_ME_CORRUPT_SCREEN, '']

    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False
    return True

def set_screensaver(cip: comms.Driver, paths: types.MEPaths, state: bool) -> bool:
    if state:
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.ENABLE_SCREENSAVER, '']
    else:
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.DISABLE_SCREENSAVER, '']

    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False
    return True

def start_process(cip: comms.Driver, paths: types.MEPaths, process: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.START_PROCESS, process]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False
    return True

def stop_process(cip: comms.Driver, paths: types.MEPaths, process: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.STOP_PROCESS, process]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False
    return True

def stop_process_me(cip: comms.Driver, paths: types.MEPaths) -> bool:
    if get_process_running(cip, paths, 'MERuntime.exe'):
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.STOP_PROCESS_ME, '']
        resp_code, resp_data = helper.run_function(cip, req_args)
        if (resp_code != 0): return False
    return True


from . import application
from . import decompress
from . import firmware
from . import fuwhelper
from . import helper
from . import messages
from . import registry
from . import transfer
from . import types
from . import util
from . import validation


from collections.abc import Callable
import os
from typing import Optional
from warnings import warn

from . import comms
from .me import firmware
from .me import fuwhelper
from .me import transfer
from .me import types
from .me import util
from .me import validation

LOCAL_RUNTIME_PATH = "C:\\Users\\Public\\Documents\\RSView Enterprise\\ME\\Runtime"
LOCAL_BIN_PATH = "C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise"
LOCAL_FUP_PATH = "C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUPs"

class MEUtility(object):
    def __init__(
        self,
        comms_path: str = None, 
        driver: str = None, 
        ignore_terminal_valid: bool = False, 
        ignore_driver_valid: bool = False,
        local_bin_path: str = None,
        local_fup_path: str = None,
        local_runtime_path: str = None,
    ):
        """
        Initializes an instance of the MEUtility class.

        Args:
            comms_path (str): The path to the communications resource (ex: 192.168.1.20).
            driver (str): The driver name to use (ex: pycomm3 or pylogix).  If not specified, will default
                the first one installed that can be found.
            ignore_terminal_valid (bool): If True, ignore terminal validation checks.
            ignore_driver_valid (bool): If True, ignore driver validation checks.
            local_bin_path (str): The default directory to assume Helper/Cover files are found.
            local_fup_path (str): The default directory to assume *.FUP files are found.
            local_runtime_path (str): The default directory to assume *.MER files are found.
        """
        self.comms_path = comms_path
        self.driver = driver
        self.ignore_terminal_valid = ignore_terminal_valid
        self.ignore_driver_valid = ignore_driver_valid

        self.local_bin_path = LOCAL_BIN_PATH if local_bin_path is None else local_bin_path
        self.local_fup_path = LOCAL_FUP_PATH if local_fup_path is None else local_fup_path 
        self.local_runtime_path = LOCAL_RUNTIME_PATH if local_runtime_path is None else local_runtime_path

    def create_firmware_card(
        self,
        fup_path_local: str,
        fwc_path_local: str,
        kep_drivers: list[str] = None,
        progress: Optional[Callable[[str, str, int, int], None]] = None
    ) -> types.MEResponse:
        """
        Creates a Firmware Card that can be used to update an ME terminal.

        Args:
            fup_path_local (str): The path to the *.FUP file to use.
            fwc_path_local (str): The path to the firmware card that will be generated (i.e. USB/CF card).
            kep_drivers (list[str]): The names of the KepDrivers to enable by default.
            progress: Optional callback for progress indication.
        """

        # Use default RSView directory if one is not specified
        if not os.path.isfile(fup_path_local):
            if os.path.sep not in fup_path_local:
                fup_path_local = os.path.join(self.local_fup_path, fup_path_local)

        # Perform firmware flash to terminal
        try:
            resp = firmware.fup_to_fwc_folder(
                input_path=fup_path_local,
                output_path=fwc_path_local,
                kep_drivers=kep_drivers,
                progress=progress
            )
        except Exception as e:
            print(e)
            return types.MEResponse(None, types.ResponseStatus.FAILURE)

        return types.MEResponse(None, types.ResponseStatus.SUCCESS)
    
    def download(
        self, 
        file_path_local: str,
        file_name_terminal: str = None,
        delete_logs: bool = False,
        overwrite: bool = False,
        replace_comms: bool = False,
        run_at_startup: bool = True,
        progress: Optional[Callable[[str, str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Downloads a *.MER file from the local device to the remote terminal.

        Args:
            file_path_local (str): The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)    
            file_name_terminal (str): If the *.MER name on the terminal should be different, specify it here.
            delete_logs (bool) : If True, will configure the terminal to delete logs atstartup.  Defaults to false.
            overwrite (bool): If True, will allow the download to overwrite an existing *.MER file on the terminal
                with the same file name.  Defaults to false.
            replace_comms (bool) : If True, will replace the terminal's communications setup with the one
                from the *.MER file being downloaded.  Defaults to false.
            run_at_startup (bool) : If True, will also set this *.MER file to be run at terminal startup and
                reboot the terminal now.  Defaults to True.
            progress: Optional callback for progress indication.
        """
        # Use default MER directory if one is not specified
        if not os.path.isfile(file_path_local):
            if os.path.sep not in file_path_local:
                file_path_local = os.path.join(self.local_runtime_path, file_path_local)

        if file_name_terminal is None: file_name_terminal = os.path.basename(file_path_local)

        with comms.Driver(self.comms_path, self.driver) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')
                
            # Validate that all starting conditions for downnload to terminal are good
            try:
                resp = validation.is_valid_download(
                    cip=cip,
                    device=self.device,
                    file_path_local=file_path_local,
                    file_name_terminal=file_name_terminal,
                    overwrite=overwrite
                )
                if resp:
                    self.device.log.append(f'Validated download for {file_path_local}.')
                else:
                    self.device.log.append(f'Failed to validate download.')
                    return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to validate download.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

            # Perform *.MER download to terminal
            try:
                resp = transfer.download_file_mer(
                    cip=cip,
                    device=self.device,
                    file_path_local=file_path_local,
                    file_name_terminal=file_name_terminal,
                    overwrite=overwrite,
                    run_at_startup=run_at_startup,
                    replace_comms=replace_comms,
                    delete_logs=delete_logs,
                    progress=progress
                )
                if not(resp):
                    self.device.log.append(f'Failed to download to terminal.')
                    return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to download to terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)
    
    def flash_firmware(
        self, 
        fup_path_local: str, 
        fuwhelper_path_local: str, 
        fuwcover_path_local: str = None,
        kep_drivers: list[str] = None,
        progress: Optional[Callable[[str, str, int, int], None]] = None
    ) -> types.MEResponse:
        """
        Flashes a firmware image to the remote terminal.

        Args:
            fup_path_local (str): The local path to the firmware file (ex: C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUPs\\ME_PVP6xX_12.00-20200922.fup)
            fuwhelper_path_local (str): The local path to the firmware helper file (ex: C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUWhelper6xX.dll)
            fuwcover_path_local (str): The local path to the firmware cover file if applicable (ex: C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUWCover4xX.exe)
            kep_drivers (list[str]): The names of the KepDrivers to enable by default.
            progress: Optional callback for progress indication.
        """
        # Use default RSView directory if one is not specified
        if not os.path.isfile(fuwhelper_path_local):
            if os.path.sep not in fuwhelper_path_local:
                fuwhelper_path_local = os.path.join(self.local_bin_path, fuwhelper_path_local)

        # Use default RSView directory if one is not specified
        if fuwcover_path_local is not None:
            if not os.path.isfile(fuwcover_path_local):
                if os.path.sep not in fuwcover_path_local:
                    fuwcover_path_local = os.path.join(self.local_bin_path, fuwcover_path_local)

        # Use default RSView directory if one is not specified
        if not os.path.isfile(fup_path_local):
            if os.path.sep not in fup_path_local:
                fup_path_local = os.path.join(self.local_fup_path, fup_path_local)

        with comms.Driver(self.comms_path, self.driver) as cip:
            if (self.driver == comms.DRIVER_NAME_PYCOMM3) and comms.is_routed_path(self.comms_path):
                if (cip._const_timeout_ticks != b'\xFF'):
                    if self.ignore_driver_valid:
                        warn('Drive pycomm3 specified with bad TIMEOUT_TICKS value, but driver validation is set to IGNORE.')
                    else:
                        resp = f"""
                            Cannot flash firmware to routed path using pycomm3 due to TIMEOUT_TICKS default value {cip._const_timeout_ticks}.
                            This will cause failures during the firmware upgrade process and require a factory reset to recover.
                            Please change the value in pycomm3.const.TIMEOUT_TICKS to b'\\xFF' to proceed, or use pylogix instead.
                        """
                        raise NotImplementedError(resp)

            # Set socket timeout first.
            # The terminal will pause at certain points and delay acknowledging messages.
            # Without this, the process will fail and the terminal will require a factory reset.
            cip.timeout = 255.0

            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)) or not(validation.is_native_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Perform firmware flash to terminal
            try:
                resp = firmware.flash_fup_to_terminal(
                    cip=cip,
                    device=self.device,
                    fup_path_local=fup_path_local,
                    fuwhelper_path_local=fuwhelper_path_local,
                    fuwcover_path_local=fuwcover_path_local,
                    kep_drivers=kep_drivers,
                    progress=progress
                )
                if not(resp):
                    self.device.log.append(f'Failed to flash terminal.')
                    return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to flash terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)

    def get_terminal_info(
        self, 
        print_log: bool = False,
        redact_log: bool = False,
        silent_mode: bool = False,
    ) -> types.MEResponse:
        """
        If no upload or download are desired, where terminal info would typically be checked
        as a prerequisite, this function can be called to generate similar log entries to
        get information about the remote terminal.

        Args:
            print_log (bool): If True, will print values.  Please include
                a copy with any bug reports!  Defaults to false.
            redact_log (bool): If True, will exclude potentially sensitive values
                such as the *.MER file names on the remote terminal. Defaults to False.
            silent_mode (bool): If True, will exclude values that require invocations
                or transfers so that Diagnostics window doesn't pop up on the remote
                terminal.  Defaults to False.
        """
        with comms.Driver(self.comms_path, self.driver) as cip:
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                util.create_log(cip, self.device, print_log, redact_log, silent_mode)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to get terminal info.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)

    def reboot(
        self
    ) -> types.MEResponse:
        """
        Reboots the remote terminal now.
        """
        with comms.Driver(self.comms_path, self.driver) as cip:
            self.device = validation.get_terminal_info(cip)

            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                util.reboot(cip, self.device)
            except Exception as e:
                self.device.log.append(f'Failed to reboot terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)

    def stop(
        self,
        fuwhelper_path_local: str,
        progress: Optional[Callable[[str, str, int, int], None]] = None
    ) -> types.MEResponse:
        """
        Stops ME Station on the terminal.

        Args:
            fuwhelper_path_local (str): The local path to the firmware helper file (ex: C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUWhelper6xX.dll)
        """
        with comms.Driver(self.comms_path, self.driver) as cip:
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)) or not(validation.is_native_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                firmware.get_or_download_fuwhelper(
                    cip=cip,
                    device=self.device,
                    fuwhelper_path_local=fuwhelper_path_local,
                    progress=progress
                )

                if fuwhelper.get_process_running(
                    cip=cip,
                    paths=self.device.me_paths,
                    process_name='MERuntime.exe'
                ):
                    # Check major rev.  v5 lacks dedicated termination function
                    if util.get_major_rev(cip, self.device) <= 5:
                        fuwhelper.stop_process(
                            cip=cip,
                            paths=self.device.me_paths,
                            process='MERuntime.exe'
                        )
                    else:
                        fuwhelper.stop_process_me(
                            cip=cip,
                            paths=self.device.me_paths
                        )
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to stop ME Station.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)

    def upload(
        self, 
        file_path_local: str, 
        file_name_terminal: str = None,
        overwrite: bool = False,
        progress: Optional[Callable[[str, str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Uploads a *.MER file from the remote terminal to the local device.

        Args:
            file_path_local (str): The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)
            file_name_terminal (str): The remote name of the *.MER file if different from the name in the local path.
            overwrite (bool) : If True, will replace the file on the local device with the uploaded
                copy from the remote terminal.  Defaults to False.
            progress: Optional callback for progress indication.
        """

        # Create local path if it doesn't exist yet
        if not(os.path.exists(os.path.dirname(file_path_local))): os.makedirs(os.path.dirname(file_path_local), exist_ok=True)

        if file_name_terminal is None: file_name_terminal = os.path.basename(file_path_local)

        with comms.Driver(self.comms_path, self.driver) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Check for existing file
            if not(overwrite) and (os.path.exists(file_path_local)):
                self.device.log.append(f'File {file_path_local} already exists.  Use overwrite=True to overwrite existing local file from the remote terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

            # Perform *.MER upload from terminal
            try:
                resp = transfer.upload_file_mer(
                    cip=cip,
                    device=self.device,
                    file_path_local=file_path_local,
                    file_name_terminal=file_name_terminal,
                    progress=progress
                )                    
                if not(resp):
                    self.device.log.append(f'Failed to upload from terminal.')
                    return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to upload from terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)

    def upload_all(
        self, 
        folder_path_local: str, 
        overwrite: bool = False,
        progress: Optional[Callable[[str, str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Uploads all *.MER files from the remote terminal to the local device.

        Args:
            folder_path_local (str) : The local path to the target directory (ex: C:\\YourFolder)
            overwrite (bool) : If True, will replace the file on the local device with the 
                uploaded copy from the remote terminal.  Defaults to False.
            progress: Optional callback for progress indication.
        """

        # Create upload folder if it doesn't exist yet
        if not(os.path.exists(folder_path_local)): os.makedirs(folder_path_local, exist_ok=True)

        with comms.Driver(self.comms_path, self.driver) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                mer_list = transfer.upload_list_mer(cip, self.device)
                mer_list = [mer for mer in mer_list if mer]
                for file_name_terminal in mer_list:
                    file_path_local = os.path.join(folder_path_local, file_name_terminal)

                    # Check for existing *.MER
                    if not(overwrite) and (os.path.exists(file_path_local)):
                        self.device.log.append(f'File {file_path_local} already exists.  Use overwrite=True to overwrite existing local file from the remote terminal.')
                        return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
                    
                    resp = transfer.upload_file_mer(
                        cip=cip,
                        device=self.device,
                        file_path_local=file_path_local,
                        file_name_terminal=file_name_terminal,
                        progress=progress
                    )
                    if not(resp):
                        self.device.log.append(f'Failed to upload from terminal.')
                        return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to upload from terminal.')
                return types.MEResponse(self.device, types.ResponseStatus.FAILURE)
        return types.MEResponse(self.device, types.ResponseStatus.SUCCESS)


from collections.abc import Callable
import configparser
import olefile
import os
import time
from typing import Optional
from warnings import warn

from .. import comms
from . import decompress
from . import fuwhelper
from . import helper
from . import transfer
from . import types
from . import util

INFORMATION_NAME = '_INFORMATION'
OTW_USE_WIN_DIR = [
    'locOSup.exe',
    'ebcbootrom.bin',
    'system.bin',
    'upgradeoptions.exe',
    'valOSpart.exe',
    'GetFreeRAM.exe',
    'PVPlus_Mozart_nkc.MCE',
    'EBCMOZ.EBC'
]

CE_BLACKLIST_FILES = [
    'atlce400.dll',
    'symbol.ttf',
    'times.ttf',
    'wingding.ttf'
]

def _create_upgrade_dat(
    version: types.MEFupUpgradeInfVersion,
    card: types.MEFupUpgradeInfCard, 
    me_files: types.MEFupMEFileListInf,
    ce: list[tuple[str, str]] = None,
    kep_drivers: list[str] = None,
    kep_size_bytes: int = None
) -> str:
    # Always use windows-style
    crlf = '\r\n'

    # There is an upgrade.dat file which needs to be created
    # from the content in the upgrade.inf file.
    isc_size_bytes = card.storage_size_bytes + me_files.info.size_on_disk_bytes
    if kep_size_bytes: isc_size_bytes += kep_size_bytes
    fields = [
        f'PLAT={version.plat}',
        f'OS={version.os}',
        f'ME={version.me}',
        f'KEP={version.kep}',
        f'MINOS={version.minos}',
        f'MAXOS={version.maxos}',
        f'ARD={version.ard}',
        f'RAM={card.ram_size_bytes}',
        f'ISC={isc_size_bytes}',
        f'FP={card.fp_size}'
    ]
    result = ';'.join(fields) + ';' + crlf

    # If KEP drivers are present
    if kep_drivers:
        kep_section = f'{crlf}[KEPdrivers]{crlf}OPTFILE=useroptions.txt;{crlf}'
        result += kep_section

    # If CE components are present
    if ce:
        cefiles = f'{crlf}[PVPCE]{crlf}'
        for (infile, outfile) in ce: cefiles += f'{infile}={outfile}{crlf}'
        cefiles += ('\x00')
        result += cefiles

    return result

def _create_user_options(kep_drivers: list[str]) -> types.MEArchive:
    user_options = 'Installed Drivers: ' + ', '.join(kep_drivers)
    data = user_options.encode('utf-16-le', errors='ignore')
    return types.MEArchive(
        name='useroptions.txt',
        data=data,
        path=['useroptions.txt'],
        size=len(data)
    )

def _get_upgrade_inf(streams: list[types.MEArchive]) -> types.MEFupUpgradeInf:
    return _deserialize_fup_upgrade_inf(util._get_stream_by_name(streams, 'upgrade.inf').data.decode('utf-8'))

def _get_mefilelist_inf(streams: list[types.MEArchive]) -> types.MEFupMEFileListInf:
    try:
        return _deserialize_fup_mefilelist_inf(util._get_stream_by_name(streams, 'MEFileList.inf').data.decode('utf-8'))
    except Exception as e:
        # v6+ don't use this at all so just enter default values
        return types.MEFupMEFileListInf(info=types.MEFupMEFileListInfInfo(me='', size_on_disk_bytes=0), mefiles=[])

def _get_upgrade_dat(
    streams: list[types.MEArchive],
    kep_drivers: list[str] = None
) -> types.MEArchive:
    upgrade_inf_data = _get_upgrade_inf(streams)
    mefilelist_inf_data = _get_mefilelist_inf(streams)

    kep_size_bytes = 0
    if kep_drivers:
        for (driver, size) in upgrade_inf_data.drivers:
            if driver in kep_drivers: kep_size_bytes += size
            if (driver == 'Overhead'): kep_size_bytes += size

    # Should really be the specified mode (FWC or OTW) because they can be different values.
    dat_file = _create_upgrade_dat(
        version=upgrade_inf_data.version,
        card = upgrade_inf_data.otw,
        me_files=mefilelist_inf_data,
        ce=upgrade_inf_data.ce,
        kep_drivers=kep_drivers,
        kep_size_bytes=kep_size_bytes
    )
    
    #data = bytearray(dat_file, encoding='utf-16-le')
    data = dat_file.encode('utf-16-le', errors='ignore')
    return types.MEArchive(
        name='Upgrade.dat',
        data=data,
        path=['Upgrade.dat'],
        size=len(data)
    )

def _deserialize_fup_mefilelist_inf(input: str) -> types.MEFupMEFileListInf:
    config = configparser.ConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read_string(input)

    info_section = config['info']
    info = types.MEFupMEFileListInfInfo(
        me=info_section.get('ME'),
        size_on_disk_bytes=info_section.getint('SizeOnDisk', 0)
    )

    me_files_section = config['MEFILES']
    me_files = []
    for key, value in me_files_section.items():
        me_files.append(key)
    
    # Return ConfigData instance
    return types.MEFupMEFileListInf(info=info, mefiles=me_files)

def _deserialize_fup_upgrade_inf(input: str) -> types.MEFupUpgradeInf:
    config = configparser.ConfigParser(allow_no_value=True, strict=False)
    config.optionxform = str
    config.read_string(input)

    # Version Header
    version_section = config['version']
    version = types.MEFupUpgradeInfVersion(
        plat=version_section.getint('Platform', 0),
        os=version_section.get('OS'),
        me=version_section.get('ME'),
        kep=version_section.get('KEP'),
        minos=version_section.get('MINOS'),
        maxos=version_section.get('MAXOS'),
        ard=version_section.getint('ARD', 0)
    )
    
    # Firmware Card
    fwc_section = config['FWC']
    fwc_files = [
        (key, value) for key, value in fwc_section.items()
        if key not in ['AddRAMSize', 'AddISCSize', 'AddFPSize']
    ]
    fwc = types.MEFupUpgradeInfCard(
        files=fwc_files,
        ram_size_bytes=fwc_section.getint('AddRAMSize', 0),
        storage_size_bytes=fwc_section.getint('AddISCSize', 0),
        fp_size=fwc_section.getint('AddFPSize', 0)
    )

    # Over-The-Wire
    otw_section = config['OTW']
    otw_files = [
        (key, value) for key, value in otw_section.items()
        if key not in ['AddRAMSize', 'AddISCSize', 'AddFPSize']
    ]
    otw = types.MEFupUpgradeInfCard(
        files=otw_files,
        ram_size_bytes=otw_section.getint('AddRAMSize', 0),
        storage_size_bytes=otw_section.getint('AddISCSize', 0),
        fp_size=otw_section.getint('AddFPSize', 0)
    )

    # Drivers
    try:
        drivers_section = config['KEPDRIVERS']
        drivers = [(key, int(value)) for key, value in drivers_section.items()]
    except:
        drivers = []

    # CE Components
    #
    # This is where the CE files will be listed out.
    #
    # Currently does not work because there are duplicates that configparser
    # doesn't like.
    try:
        ce_section = config['PVPCE']
        ce = [(key, value) for key, value in ce_section.items()]
    except:
        ce = []

    return types.MEFupUpgradeInf(
        version=version,
        fwc=fwc,
        otw=otw,
        drivers=drivers,
        ce=ce
    )

def fup_to_fuc(
    input_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[types.MEArchive]:
    # Application-specific handling for *.FUP files that
    # keeps streams in memory.
    #
    # This results in an intermediate form that can be used to
    # form the firmware upgrade card or over-the-wire format.
    with olefile.OleFileIO(input_path) as ole:
        streams = decompress.decompress_archive(
            ole=ole,
            progress=progress
        )

        # In the *.FUP packages specifically, there is some data
        # in the upgrade.inf file that needs to be arranged into
        # an Upgrade.dat file.
        streams.insert(0, _get_upgrade_dat(streams=streams, kep_drivers=kep_drivers))

        # If KepDrivers were specified, insert the useroptions.txt
        # file to enumerate them
        if kep_drivers: streams.insert(0, _create_user_options(kep_drivers=kep_drivers))
        return streams

def fup_to_fuc_folder(
    input_path: str,
    output_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None,
):
    # Application-specific handling for *.FUP files that
    # writes streams to a folder.
    #
    # This results in an intermediate form that can be used to
    # form the firmware upgrade card or over-the-wire format.

    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)
    streams = fup_to_fuc(
        input_path=input_path,
        kep_drivers=kep_drivers,
        progress=progress
    )
    for stream in streams:
        # In *.FUP packages specifically, there are some _INFORMATION files
        # that don't need to be exported
        if stream.name.endswith(INFORMATION_NAME): continue

        stream_output_path = decompress._create_subfolders(output_path, stream.path)
        with open(stream_output_path, 'wb') as f:
            f.write(stream.data)

def fup_to_fwc(
    input_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None,
) -> list[types.MEArchive]:
    # Application-specific handling for *.FUP files that
    # keeps streams in memory.
    #
    # This results in the Firmware Card format (FWC) that
    # can be used to flash a terminal via removable media.
    streams = fup_to_fuc(
        input_path=input_path,
        kep_drivers=kep_drivers,
        progress=progress
    )
    upgrade_inf = _get_upgrade_inf(streams)
    
    streams_fwc = []
    for (file, outfile) in upgrade_inf.fwc.files:
        stream = util._get_stream_by_name(streams, file)
        stream.path = util._path_to_list(outfile)
        streams_fwc.append(stream)

    for (file, outfile) in upgrade_inf.ce:
        dirname, basename = util.split_file_path(outfile)
        stream = util._get_stream_by_name(streams, file)
        stream.path = ['upgrade', 'AddIns', basename]
        streams_fwc.append(stream)

    if kep_drivers:
        stream = util._get_stream_by_name(streams, 'useroptions.txt')
        stream.path = ['upgrade', 'useroptions.txt']
        streams_fwc.insert(0, stream)

    return streams_fwc

def fup_to_fwc_folder(
    input_path: str,
    output_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None,
):
    # Application-specific handling for *.FUP files that
    # writes streams to a folder.
    #
    # This results in the Firmware Card format (FWC) that
    # can be used to flash a terminal via removable media.

    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)
    streams = fup_to_fwc(
        input_path=input_path,
        kep_drivers=kep_drivers,
        progress=progress
    )
    for stream in streams:
        stream_output_path = decompress._create_subfolders(output_path, stream.path)
        with open(stream_output_path, 'wb') as f:
            f.write(stream.data)

def fup_to_otw(
    input_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[types.MEArchive]:
    # Application-specific handling for *.FUP files that
    # keeps streams in memory.
    #
    # This results in the Over-The-Wire format (OTW) that
    # can be sent via a network connection.
    streams = fup_to_fuc(
        input_path=input_path,
        kep_drivers=kep_drivers,
        progress=progress
    )
    upgrade_inf = _get_upgrade_inf(streams)

    streams_otw = []
    for (file, outfile) in upgrade_inf.otw.files:
        stream = util._get_stream_by_name(streams, file)
        stream.path = util._path_to_list(outfile)
        streams_otw.append(stream)

    for (file, outfile) in upgrade_inf.ce:
        stream = util._get_stream_by_name(streams, file)
        stream.path = util._path_to_list(outfile)
        streams_otw.append(stream)

    if kep_drivers:
        stream = util._get_stream_by_name(streams, 'useroptions.txt')
        stream.path = ['upgrade', 'useroptions.txt']
        streams_otw.insert(0, stream)

    return streams_otw

def fup_to_otw_folder(
    input_path: str,
    output_path: str,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None,
):
    # Application-specific handling for *.FUP files that
    # writes streams to a folder.
    #
    # This results in the Over-The-Wire format (OTW) that
    # can be sent via a network connection.  Note that OTW
    # is not usually stored in a folder so this is intended
    # just for debug.

    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)
    streams = fup_to_otw(
        input_path=input_path,
        kep_drivers=kep_drivers,
        progress=progress
    )
    for stream in streams:
        stream_output_path = decompress._create_subfolders(output_path, stream.path)
        with open(stream_output_path, 'wb') as f:
            f.write(stream.data)

def get_or_download_fuwhelper(
    cip: comms.Driver,
    device: types.MEDeviceInfo,
    fuwhelper_path_local: str,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    # Determine if firmware upgrade helper already exists in one
    # of the expected locations and use it, or else transfer the
    # helper file specified.
    if helper.get_file_exists(cip, device.me_paths, '\\Windows\\FUWhelper.dll'):
        device.me_paths.fuwhelper_file = '\\Windows\\FUWhelper.dll'
    elif helper.get_file_exists(cip, device.me_paths, device.me_paths.fuwhelper_file):
        pass
    else:
        transfer.download_file(
            cip=cip,
            device=device,
            file_path_local=fuwhelper_path_local,
            file_path_terminal=device.me_paths.fuwhelper_file,
            overwrite=True,
            progress=progress
        )
        util.wait(time_sec=5, progress=progress)

def flash_fup_to_terminal(
    cip: comms.Driver, 
    device: types.MEDeviceInfo,
    fup_path_local: str,
    fuwhelper_path_local: str,
    fuwcover_path_local: str = None,
    kep_drivers: list[str] = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    # Read FUP into memory
    streams_otw = fup_to_otw(
        input_path=fup_path_local,
        kep_drivers=kep_drivers,
        progress=progress
    )

    # Ensure firmware upgrade helper is in place
    get_or_download_fuwhelper(
        cip=cip,
        device=device,
        fuwhelper_path_local=fuwhelper_path_local,
        progress=progress
    )

    if util.get_major_rev(cip, device) <= 5:
        mefilelist_inf_data = _get_mefilelist_inf(streams_otw)

        fuwhelper.set_screensaver(cip, device.me_paths, False)
        fuwhelper.set_me_corrupt_screen(cip, device.me_paths, False)
        os_rev = fuwhelper.get_os_rev(cip, device.me_paths)
        part_size = fuwhelper.get_partition_size(cip, device.me_paths)
        restore = fuwhelper.get_file_exists(cip, device.me_paths, '\\Storage Card\\_restore_reserve.cmd')
        fuwhelper.start_process(cip, device.me_paths, 'GenReserve:0')

        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage_Card')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\upgrade')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card\\upgrade')
        fuwhelper.clear_folder(cip, device.me_paths, '\\Storage Card\\upgrade')

        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Windows')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Windows')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Windows\\upgrade')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Windows\\upgrade')
        fuwhelper.clear_folder(cip, device.me_paths, '\\Windows\\upgrade')

        if (fuwhelper.get_file_exists(cip, device.me_paths, '\\Storage Card\\Step2.dat')):
            try:
                fuwhelper.delete_file(cip, device.me_paths, '\\Storage Card\\Step2.dat')
            except Exception as e:
                print(e)

        storage_free_space = fuwhelper.get_free_space(cip, device.me_paths, '\\Storage Card')
        storage_total_space = fuwhelper.get_total_space(cip, device.me_paths, '\\Storage Card')
        windows_free_space = fuwhelper.get_free_space(cip, device.me_paths, '\\Windows')
        windows_total_space = fuwhelper.get_total_space(cip, device.me_paths, '\\Windows')

        # Check files from MEFileInfo.inf?
        for file in mefilelist_inf_data.mefiles:
            fuwhelper.get_file_exists(cip, device.me_paths, f'\\Storage Card{file}')

        transfer.download_file(
            cip=cip,
            device=device,
            file_path_local=fuwcover_path_local,
            file_path_terminal='\\Windows\\FUWCover.exe',
            overwrite=True,
            progress=progress
        )
        fuwhelper.start_process(cip, device.me_paths, '\\Windows\\FUWCover.exe')
        fuwhelper.stop_process(cip, device.me_paths, 'MERuntime.exe')
        fuwhelper.clear_folder(cip, device.me_paths, '\\Storage Card\\Rockwell Software\\RSViewME')

        # Delete files from MEFileInfo.inf?
        for file in mefilelist_inf_data.mefiles:
            if fuwhelper.get_file_exists(cip, device.me_paths, f'\\Storage Card{file}'):
                try:
                    fuwhelper.delete_file(cip, device.me_paths, f'\\Storage Card{file}')
                except Exception as e:
                    print(e)

        # Delete KEPServer
        if fuwhelper.get_file_exists(cip, device.me_paths, '\\Storage Card\\KEPServerEnterprise\\Uninstall KepServerEnterprise.exe'):
            try:
                fuwhelper.start_process(cip, device.me_paths, '\\Storage Card\\KEPServerEnterprise\\Uninstall KepServerEnterprise.exe:/S')
            except Exception as e:
                print(e)
        if fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\KEPServerEnterprise'):
            # Interesting note on ClearRemDirectory... maybe the ? at the end is a best-effort and suppresses
            # failures?
            try:
                fuwhelper.clear_folder(cip, device.me_paths, '\\Storage Card\\KEPServerEnterprise?')
            except Exception as e:
                print(e)
            try:
                fuwhelper.delete_folder(cip, device.me_paths, '\\Storage Card\\KEPServerEnterprise')
            except Exception as e:
                print(e)

        for stream in streams_otw:
            # Certain CE files fail to download, still investigating
            if stream.name.lower() in CE_BLACKLIST_FILES: continue

            # No need to transfer Kepware install if no drivers specified
            if (not kep_drivers) and (stream.path[-1].lower() == 'kepwareceinstall.exe'): continue

            if stream.path[0].lower() != 'windows' and stream.path[0].lower() != 'storage card':
                # The normal files look to all have relative directories.
                #
                # Some go to \\Storage Card, others to \\Windows.
                # There is no hint in upgrade.inf file for this.
                #
                # Current guesses...
                # [1] Always redirect files after Autoapp.bat and before RFOn.bat?
                # [2] Some way to parse contents of autoapp.bat to see which are referenced?
                # [3] All binary/executable files except for known ones that belong to Storage Card?
                # [4] There is no logic to it.
                if stream.path[-1].lower() in [f.lower() for f in OTW_USE_WIN_DIR]:
                    stream_path_terminal = '\\Windows\\' + '\\'.join(stream.path)
                else:
                    stream_path_terminal = '\\Storage Card\\' + '\\'.join(stream.path)
            else:
                # The CE addon files look to all have absolute directories
                stream_path_terminal = '\\' + '\\'.join(stream.path)

            try:
                # Currently blocking this around TRY/EXCEPT just to see what happens on terminal.
                # Remove prior to release.
                warn('Still ignoring download failure for CE test!')
                transfer.download(
                    cip=cip,
                    device=device,
                    file_data=stream.data,
                    file_path_terminal=stream_path_terminal,
                    overwrite=True,
                    progress=progress
                )
            except Exception as e:
                print(e)

        # Initiate install
        fuwhelper.set_screensaver(cip, device.me_paths, True)
        fuwhelper.set_me_corrupt_screen(cip, device.me_paths, True)
        util.wait(time_sec=5, progress=progress)
        fuwhelper.stop_process(cip, device.me_paths, 'FUWCover.exe')
        fuwhelper.start_process(cip, device.me_paths, '\\Storage Card\\upgrade\\autorun.exe')
    else:
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\vfs')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card\\vfs')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\vfs\\platform firmware')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card\\vfs\\platform firmware')
        if (fuwhelper.get_file_exists(cip, device.me_paths, '\\Storage Card\\Step2.dat')):
            fuwhelper.delete_file(cip, device.me_paths, '\\Storage Card\\Step2.dat')
        if fuwhelper.get_process_running(cip, device.me_paths, 'MERuntime.exe'):
            fuwhelper.stop_process_me(cip, device.me_paths)
        if not kep_drivers:
            if fuwhelper.get_file_exists(cip, device.me_paths, '\\Windows\\useroptions.txt'):
                fuwhelper.delete_file(cip, device.me_paths, '\\Windows\\useroptions.txt')

        for stream in streams_otw:
            if stream.name == 'useroptions.txt': stream.path = ['Windows', 'useroptions.txt']
            if stream.path[0].lower() != 'windows' and stream.path[0].lower() != 'storage card' and stream.path[0].lower() != 'vfs':
                # Files with relative directories will be redirected to Storage Card
                stream_path_terminal = '\\Storage Card\\' + '\\'.join(stream.path)
            else:
                # Files with absolute directories
                stream_path_terminal = '\\' + '\\'.join(stream.path)

            transfer.download(
                cip=cip,
                device=device,
                file_data=stream.data,
                file_path_terminal=stream_path_terminal,
                overwrite=True,
                progress=progress
            )

    return True


from collections.abc import Callable
import os
import time
from typing import Optional
import zlib

from .. import comms
from . import helper
from . import registry
from . import transfer
from . import types

def crc32_checksum(data):
    return zlib.crc32(data) & 0xFFFFFFFF

def create_log(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    print_log: bool, 
    redact_log: bool, 
    silent_mode: bool
):
    if print_log: print(f'Terminal CIP identity: {device.cip_identity}.')
    if print_log: print(f'Terminal ME identity: {device.me_identity}.')
    if print_log: print(f'Terminal ME paths: {device.me_paths}.')

    try:
        line = f'Terminal has {helper.get_free_space_runtime(cip, device.me_paths)} free bytes.'
    except:
        line = f'Failed to get free space on terminal.'
    device.log.append(line)
    if print_log: print(f'{line}')

    if not silent_mode:
        try:
            files = transfer.upload_list_med(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MED files: {files}.'
            if len(files) > 0: device.running_med_file = files[0]
        except:
            line = f'Failed to list MED files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

        try:
            files = transfer.upload_list_mer(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MER files: {files}.'
            device.files = files
        except:
            line = f'Failed to list MER files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

    try:
        file = registry.get_startup_mer(cip)
        if redact_log: file = 'Redacted'
        line = f'Terminal startup file: {file}.'
        if file.lower().endswith('.mer'): device.startup_mer_file = file.split('\\')[-1]
    except:
        if get_major_rev(cip, device) <= 5:
            # For PanelView Plus 5.10 and earlier this registry key appears to be unavailable.
            line = f'Terminal startup file: could not be determined due to hardware version.'
        else:
            # If no startup app has been defined, it will also fail. 
            line = f'Terminal startup file: not configured.'
    device.log.append(line)
    if print_log: print(f'{line}')

def get_major_rev(
    cip: comms.Driver,
    device: types.MEDeviceInfo
) -> int:
    major_rev = int(device.me_identity.me_version.split(".")[0])
    return major_rev

def _get_stream_by_name(streams: list[types.MEArchive], name: str, case_insensitive=True) -> types.MEArchive:
    if case_insensitive:
        return next(x for x in streams if x.name.lower() == name.lower())
    else:
        return next(x for x in streams if x.name == name)
    
def _path_to_list(path: str) -> list[str]:
    path = path.replace('\\', '/').lower()
    components = [comp for comp in path.split('/') if comp]
    return components if components else [path]

def reboot(
    cip: comms.Driver, 
    device: types.MEDeviceInfo
):
    cip1 = comms.Driver(cip._original_path)
    cip1.timeout = 0.25
    cip1.open()
    try:
        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.me_paths)

        # If we made it here... the reboot function didn't throw
        # an exception, which means it didn't reboot.
        #
        # Further investigation needed.
        device.log.append(f'Initial reboot unsuccessful.')

        # Currently tailoring additional action to terminals where:
        #
        # Firmware version is 12.108+
        # Application is set to load at startup
        # ME Station is set to run application at startup
        # Current App is defined
        if ((device.me_identity.major_rev < 12) or ((device.me_identity.major_rev == 12) and (device.me_identity.minor_rev < 108))):
            device.log.append(f'Did not attempt additional reboot because terminal is below minimum applicable version.')
            return
        if (registry.get_startup_options(cip) != 1):
            device.log.append(f'Did not attempt additional reboot because terminal Startup Options are not set to Run Current Application.')
            return
        startup_file = os.path.basename(registry.get_startup_mer(cip).replace('\\','/'))
        if not(startup_file.lower().endswith('.mer')):
            device.log.append(f'Did not attempt additional reboot because terminal does not have valid startup *.MER defined.')
            return

        # Rebuild existing ME Startup Shortcut on terminal
        delete_logs = registry.get_startup_delete_logs(cip)
        replace_comms = registry.get_startup_replace_comms(cip)
        device.log.append(f'Setting file: {startup_file} to run at startup with Replace Comms: {replace_comms}, Delete Logs: {delete_logs}.')
        helper.create_me_shortcut(cip, device.me_paths, startup_file, replace_comms, delete_logs)

        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.me_paths)
    except Exception as e:
        # Unlike most CIP messages, this one is expected to
        # create an exception.  When it is received by the terminal,
        # the device reboots and breaks the socket.
        if (str(e) != 'failed to receive reply'): raise e
    
    cip1.close()

def split_file_path(file_path_terminal: str) -> tuple[str, str]:
    dirname, basename = file_path_terminal.rsplit('\\', 1)
    return dirname, basename

def wait(
    time_sec: int,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    elapsed = 0
    if time_sec < 0: time_sec = 1
    while elapsed < time_sec:
        elapsed += 1
        time.sleep(1)
        if progress:
            progress('Waiting', 'seconds', time_sec, elapsed)


from .. import comms
from ..common.types import CIPMessagePreset

CFFilePreamble = CIPMessagePreset(
    service=int.from_bytes(b'\x51', byteorder='big'),
    class_code=int.from_bytes(b'\xA1', byteorder='big'),
    instance=None,
    attribute=None
)

CFFileChunk = CIPMessagePreset(
    service=int.from_bytes(b'\x52', byteorder='big'),
    class_code=int.from_bytes(b'\xA1', byteorder='big'),
    instance=None,
    attribute=None
)

def dmk_preamble(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=CFFilePreamble.service,
        class_code=CFFilePreamble.class_code,
        instance=instance,
        attribute=CFFilePreamble.attribute,
        request_data=data,
        connected=True
    )

def dmk_chunk(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=CFFileChunk.service,
        class_code=CFFileChunk.class_code,
        instance=instance,
        attribute=CFFileChunk.attribute,
        request_data=data,
        connected=True
    )


from dataclasses import dataclass, field

from ..common.types import CIPIdentity, ResponseStatus

@dataclass
class DMKContentHeader:
    dmk_type: str
    family_name: str
    revision: str
    number_catalogs: int

@dataclass
class DMKContentCatalog:
    catalog: str
    revision: str
    vendor_id: int
    vendor_id_mask: int
    product_type: int
    product_type_mask: int
    product_code: int
    product_code_mask: int
    device_instance: int
    disable_sub_minor: int
    driver_name: str

@dataclass
class DMKContentFile:
    header: DMKContentHeader
    catalogs: list[DMKContentCatalog]

@dataclass
class DMKNvsHeader:
    description: str
    programming_protocol: str
    disable_sub_minor: int
    new_revision: str
    number_updates: int
    connection_type: str
    max_power_up_seconds: int
    number_identities: int
    family_name: str
    release_notes_file_name: str

@dataclass
class DMKNvsUpdate:
    nvs_instance: int
    major_revision: int
    minor_revision: int
    tftp_timeout_seconds: int
    max_timeout_seconds: int
    starting_location: int
    file_size: int
    data_file_name: str
    rename_data_file: str
    update_reset: int
    auto_reset_on_error: int
    first_transfer_delay: int
    last_transfer_delay: int
    error_instructions: str

@dataclass
class DMKNvsFile:
    header: DMKNvsHeader
    updates: list[DMKNvsUpdate]

@dataclass
class DMKFile:
    content: DMKContentFile
    nvs: DMKNvsFile

@dataclass
class CFDeviceInfo:
    comms_path: str
    cip_identity: CIPIdentity
    log: list[str]

@dataclass
class CFResponse(object):
    device: CFDeviceInfo
    status: str


from collections.abc import Callable
from enum import IntEnum
import os
import struct
from typing import Optional
from warnings import warn

from .. import comms
from . import messages
from . import helper
from . import types
from . import util

END_OF_FILE = b'\x00\x00\x00\x00\x02\x00\xff\xff'

# Known values associated with some file services, with unclear purpose or meaning.
# Further investigation needed.
GET_UNK1_VALUES = {
    b'\x02\x00'
}
GET_UNK2_VALUES = {
    b'\x01\x60',
    b'\x03\x41'
}
GET_UNK3_VALUES = {
    b'\x64\x00'
}

class TransferType(IntEnum):
    DOWNLOAD = int.from_bytes(b'\x01', byteorder='big')
    UPLOAD = int.from_bytes(b'\x00', byteorder='big')

def _create_download(
    cip: comms.Driver, 
    file_path_terminal: str,
    file_size: int,
    overwrite: bool = True,
) -> int:
    '''
    Creates a transfer instance for downloading from the local device to the remote terminal.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Byte 0        | Transfer Type (always 1 for file download)          |
        | Byte 1        | Overwrite flag (0 = New File, 1 = Overwrite)        |
        | Bytes 2->3    | Chunk size in bytes                                 |
        | Bytes 4->7    | File size in bytes                                  |
        | Bytes 8->N-1  | File name                                           |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->1    | Message instance (should match request, 0x00)       |
        | Bytes 2->3    | Unknown purpose                                     |
        | Bytes 4->5    | Transfer instance (use this instance for download)  |
        | Bytes 6->7    | Chunk size in bytes                                 |
    '''
    req_header = struct.pack('<BBHI', TransferType.DOWNLOAD, int(overwrite), cip.me_chunk_size, file_size)
    req_args = [file_path_terminal]
    req_data = req_header + b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.create_transfer(cip, req_data)
    resp_exception_text = f'Failed to create transfer instance for download of {file_path_terminal}.'
    if not resp: raise Exception(f'{resp_exception_text}.  No message response.')
 
    resp_msg_instance, resp_unk1, resp_transfer_instance, resp_chunk_size = struct.unpack('<HHHH', resp.value)
    if (resp_msg_instance != 0): raise Exception(f'{resp_exception_text}.  Response message instance: {resp_msg_instance}, expected: 0.  There may be an incomplete transfer.  Reboot terminal and try again.')
    if (resp_unk1 != 0 ): raise Exception(f'{resp_exception_text}.  Response UNK1 bytes: {resp_unk1}, expected: 0.  Please file a bug report with all available information.')
    if (resp_chunk_size != cip.me_chunk_size): raise Exception(f'{resp_exception_text}.  Response chunk size: {resp_chunk_size}, expected: {cip.me_chunk_size}.  Please file a bug report with all available information.')
    return resp_transfer_instance

def _create_upload(
    cip: comms.Driver, 
    file_path_terminal: str
) -> tuple[int, int]:
    '''
    Creates a transfer instance for uploading from the remote terminal to the local device.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Byte 0        | Transfer Type (always 0 for file upload)            |
        | Byte 1        | Unknown purpose (overwrite in download, N/A?)       |
        | Bytes 2->3    | Chunk size in bytes                                 |
        | Bytes 4->N-1  | File name                                           |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->1    | Message instance (should match request, 0x00)       |
        | Bytes 2->3    | Unknown purpose                                     |
        | Bytes 4->5    | Transfer instance (use this instance for upload)    |
        | Bytes 6->7    | Chunk size in bytes                                 |
        | Bytes 8->11   | File size in bytes                                  |
    '''
    req_header = struct.pack('<BBH', TransferType.UPLOAD, 0x00, cip.me_chunk_size)
    req_args = [f'{file_path_terminal}']
    req_data = req_header + b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.create_transfer(cip, req_data)
    resp_exception_text = f'Failed to create transfer instance for upload of {file_path_terminal}.'
    if not resp: raise Exception(f'{resp_exception_text}.  No message response.')

    resp_msg_instance, resp_unk1, resp_transfer_instance, resp_chunk_size, resp_file_size = struct.unpack('<HHHHI', resp.value)
    if (resp_msg_instance != 0): raise Exception(f'{resp_exception_text}.  Response message instance: {resp_msg_instance}, expected: 0.  There may be an incomplete transfer.  Reboot terminal and try again.')
    if (resp_unk1 != 0 ): raise Exception(f'{resp_exception_text}.  Response UNK1 bytes: {resp_unk1}, expected: 0.  Please file a bug report with all available information.')
    if (resp_chunk_size != cip.me_chunk_size): raise Exception(f'{resp_exception_text}.  Response chunk size: {resp_chunk_size}, expected: {cip.me_chunk_size}.  Please file a bug report with all available information.')
    return resp_transfer_instance, resp_file_size

def _delete(cip: comms.Driver, instance: int):
    return messages.delete_transfer(cip, instance)

def _write_download(
    cip: comms.Driver, 
    file_data: bytearray, 
    instance: int, 
    progress_desc: str = None, 
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bool:
    """
    Downloads a file from the local device to the remote terminal.
    The transfer happens by breaking the file down into one or more
    chunks, with each chunk being sent via a CIP message and reassembled
    on the remote terminal.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Chunk number                                        |
        | Bytes 4->5    | Chunk size in bytes                                 |
        | Bytes 6->N    | Chunk data                                          |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Unknown purpose                                     |
        | Bytes 4->7    | Chunk number echo                                   |
        | Bytes 8->11   | Next chunk number expected                          |
    """
    req_chunk_number = 1
    req_offset = 0
    total_bytes = len(file_data)
    while req_offset < total_bytes:
        req_chunk = file_data[req_offset:req_offset + cip.me_chunk_size]
        req_header = struct.pack('<IH', req_chunk_number, len(req_chunk))
        req_next_chunk_number = req_chunk_number + 1
        req_data = req_header + req_chunk

        # End of file
        if not req_chunk: break

        resp = messages.write_file_chunk(cip, instance, req_data)
        if not resp: raise Exception(f'Failed to write chunk {req_chunk_number} to terminal.')
        resp_unk1, resp_chunk_number, resp_next_chunk_number = struct.unpack('<III', resp.value)
        if (resp_unk1 != 0 ): raise Exception(f'Response unknown bytes: {resp_unk1}, expected: 0.')
        if (resp_chunk_number != req_chunk_number ): raise Exception(f'Response chunk number: {resp_chunk_number}, expected: {req_chunk_number}.')
        if (resp_next_chunk_number != req_next_chunk_number): raise Exception(f'Response next chunk number: {resp_next_chunk_number}, expected: {req_next_chunk_number}.')

        # Update progress callback
        current_bytes = req_offset + len(req_chunk)
        if progress: progress(f'Download {progress_desc}','bytes', total_bytes, current_bytes)

        # Continue to next chunk
        req_chunk_number += 1
        req_offset += len(req_chunk)

    # Close out file
    req_data = END_OF_FILE
    resp = messages.write_file_chunk(cip, instance, req_data)
    return True

def _read_upload(
    cip: comms.Driver, 
    file_size: int, 
    instance: int, 
    progress_desc: str = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bytearray:
    """
    Uploads a file from the remote terminal to the local device.
    The transfer happens by breaking the file down into one or more
    chunks, with each chunk being sent via a CIP message and reassembled
    on the local device.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Chunk number                                        |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Unknown purpose                                     |
        | Bytes 4->7    | Chunk number echo                                   |
        | Bytes 8->9    | Chunk size in bytes                                 |
        | Bytes 10->N   | Chunk data                                          |
    """
    req_chunk_number = 1
    resp_binary = bytearray()
    while True:
        req_data = struct.pack('<I', req_chunk_number)

        resp = messages.read_file_chunk(cip, instance, req_data)
        if not resp: raise Exception(f'Failed to read chunk {req_chunk_number} to terminal.')
        resp_unk1 = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
        resp_chunk_number = int.from_bytes(resp.value[4:8], byteorder='little', signed=False)
        resp_chunk_size = int.from_bytes(resp.value[8:9], byteorder='little', signed=False)
        resp_data = bytes(resp.value[10:])

        if (resp_unk1 != 0 ): raise Exception(f'Response unknown bytes: {resp_unk1}, expected: 0.')
        if (resp_chunk_number != req_chunk_number) and (resp_chunk_number != 0): raise Exception(f'Response chunk number: {resp_chunk_number}. expected: {req_chunk_number}.')

        # End of file
        if (resp_chunk_number == 0) and (resp_chunk_size == 2) and (resp_data == b'\xff\xff'):
            #print('End of file')
            break

        # Write to mer list
        resp_binary += resp_data

        # Update progress callback
        if progress: progress(f'Upload {progress_desc}','bytes', file_size,len(resp_binary))

        # Continue to next chunk
        req_chunk_number += 1

    return resp_binary

def _is_ready(cip: comms.Driver) -> bool:
    # I don't know what any of these three attributes are for yet.
    # It may be checking that the file exchange is available.
    resp = messages.read_file_ready(cip, b'\x30\x01')
    if not resp: return False
    if resp.value not in GET_UNK1_VALUES:
        warn(f'Invalid UNK1 value.  Please file a bug report with all available information.')
        return False

    resp = messages.read_file_ready(cip, b'\x30\x08')
    if not resp: return False
    if resp.value not in GET_UNK2_VALUES:
        warn(f'Invalid UNK2 value.  Please file a bug report with all available information.')
        return False

    resp = messages.read_file_ready(cip, b'\x30\x09')
    if not resp: return False
    if resp.value not in GET_UNK3_VALUES:
        warn(f'Invalid UNK3 value.  Please file a bug report with all available information.')
        return False

    return True

def _set_ready(cip: comms.Driver) -> bool:
    # I don't know what setting this attribute does yet.
    # It may be marking the file exchange as in use.
    #
    resp = messages.write_file_ready(cip, b'\x30\x01\xff\xff')
    if not resp: return False

    return True

def download(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_data: bytearray, 
    file_path_terminal: str, 
    overwrite: bool = False,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bool:
    instance = None
    try:
        dirname, basename = util.split_file_path(file_path_terminal)
        try:
            # Attempt to ensure the directory exists
            # Still having trouble with some directories
            helper.create_folders(cip, device.me_paths, dirname)
        except Exception as e:
            print(e)

        file_exists = helper.get_file_exists(cip, device.me_paths, file_path_terminal)
        if (overwrite and not file_exists): overwrite = False
        if (file_exists and not overwrite): raise FileExistsError(f'File {file_path_terminal} exists on terminal already and overwrite was not specified.')

        if not(_is_ready(cip)): raise Exception('Terminal not ready for file transfer lock.')
        instance = _create_download(
            cip=cip,
            file_path_terminal=file_path_terminal,
            file_size=len(file_data),
            overwrite=overwrite
        )

        if not(_set_ready(cip)): raise Exception('Terminal refused file transfer lock.')
        _write_download(
            cip=cip,
            file_data=file_data,
            instance=instance,
            progress_desc=file_path_terminal,
            progress=progress
        )
        device.log.append(f'Downloaded {file_path_terminal} using transfer instance {instance}.')

        _delete(
            cip=cip,
            instance=instance
        )
    except Exception as e:
        if instance is not None: _delete(cip=cip, instance=instance)
        raise Exception(f'Download {file_path_terminal} failed: {str(e)}')

    return True

def download_file(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_path_local: str,
    file_path_terminal: str,
    overwrite: bool = True,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bool:
    with open(file_path_local, 'rb') as source_file:
        return download(
            cip=cip,
            device=device,
            file_data=bytearray(source_file.read()),
            file_path_terminal=file_path_terminal,
            overwrite=overwrite,
            progress=progress
        )
    
def download_file_mer(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_path_local: str,
    file_name_terminal: str,
    overwrite: bool,
    run_at_startup: bool,
    replace_comms: bool,
    delete_logs: bool,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bool:
    file_path_terminal = f'{device.me_paths.runtime}\\{file_name_terminal}'
    download_file(
        cip=cip,
        device=device,
        file_path_local=file_path_local,
        file_path_terminal=file_path_terminal,
        overwrite=overwrite,
        progress=progress
    )
    if run_at_startup:
        helper.create_me_shortcut(
            cip=cip,
            paths=device.me_paths,
            file=file_name_terminal,
            replace_comms=replace_comms,
            delete_logs=delete_logs
        )
        util.reboot(cip, device)
    return True

def upload(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_path_terminal: str, 
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bytearray:
    instance = None
    try:
        file_exists = helper.get_file_exists(cip=cip, paths=device.me_paths, file_path=file_path_terminal)
        if file_exists:
            instance, file_size = _create_upload(cip=cip, file_path_terminal=file_path_terminal)
            resp_binary = _read_upload(
                cip=cip,
                file_size=file_size,
                instance=instance,
                progress_desc=file_path_terminal,
                progress=progress
            )
            _delete(cip=cip, instance=instance)
            return resp_binary
        raise FileNotFoundError(f'File {file_path_terminal} does not exist on terminal.')
    except Exception as e:
        if instance is not None: _delete(cip=cip, instance=instance)
        raise Exception(f'Upload {file_path_terminal} failed: {str(e)}')

def upload_file(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_path_local: str,
    file_path_terminal: str,
    progress: Optional[Callable[[str, str, int, int], None]] = None
):
    resp_binary = upload(
        cip=cip,
        device=device,
        file_path_terminal=file_path_terminal,
        progress=progress
    )
    if not(os.path.exists(file_path_local)): os.makedirs(os.path.dirname(file_path_local), exist_ok=True)
    with open(file_path_local, 'wb') as dest_file:
        dest_file.write(resp_binary)

def upload_file_mer(
    cip: comms.Driver, 
    device: types.MEDeviceInfo, 
    file_path_local,
    file_name_terminal,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bool:
    file_path_terminal = f'{device.me_paths.runtime}\\{file_name_terminal}'
    upload_file(
        cip=cip,
        device=device,
        file_path_local=file_path_local,
        file_path_terminal=file_path_terminal,
        progress=progress
    )
    return True

def upload_list(
    cip: comms.Driver,
    device: types.MEDeviceInfo,
    file_path_terminal: str
) -> list[str]:
    resp_binary = upload(
        cip=cip,
        device=device,
        file_path_terminal=file_path_terminal,
        progress=None
    )
    resp_str = "".join([chr(b) for b in resp_binary if b != 0])
    resp_list = resp_str.split(':')
    return resp_list

def upload_list_med(
    cip: comms.Driver, 
    device: types.MEDeviceInfo
) -> list[str]:
    try:
        helper.create_file_list_med(cip, device.me_paths)
        file_list = upload_list(
            cip=cip,
            device=device,
            file_path_terminal=f'{device.me_paths.upload_list}'
        )
        helper.delete_file_list(cip, device.me_paths)        
        return file_list
    except Exception as e:
        return None
    
def upload_list_mer(
    cip: comms.Driver, 
    device: types.MEDeviceInfo
) -> list[str]:
    try:
        helper.create_file_list_mer(cip, device.me_paths)
        file_list = upload_list(
            cip=cip,
            device=device,
            file_path_terminal=f'{device.me_paths.upload_list}'
        )
        helper.delete_file_list(cip, device.me_paths)
        return file_list
    except Exception as e:
        return None


from dataclasses import dataclass, field
from enum import StrEnum

@dataclass
class CIPIdentity:
    hardware_rev: int
    major_rev: int
    minor_rev: int
    product_code: int
    product_name: str
    product_type: int
    serial_number: str = field(repr=False)
    status: int
    vendor_id: int

@dataclass
class CIPMessagePreset:
    service: int
    class_code: int
    instance: int
    attribute: int

class ResponseStatus(StrEnum):
    SUCCESS = 'Success'
    FAILURE = 'Failure'


import struct

from .. import comms
from . import messages
from . import types

def get_cip_identity(cip: comms.Driver) -> types.CIPIdentity:
    resp1 = messages.get_identity(cip)
    vendor_id, product_type, product_code, major_rev, minor_rev, status, serial_number, product_name_length = struct.unpack('<HHHBBHLB', resp1.value[:15])
    product_name = resp1.value[15:15 + product_name_length].decode('utf-8', errors='ignore')
    serial_number_str = f'{serial_number:08x}'

    try:
        resp2 = messages.get_hardware_rev(cip)
        hardware_rev = struct.unpack('<H', resp2.value)[0]
    except:
        hardware_rev = None

    return types.CIPIdentity(
        hardware_rev=hardware_rev,
        major_rev=major_rev,
        minor_rev=minor_rev,
        product_code=product_code,
        product_name=product_name,
        product_type=product_type,
        serial_number=serial_number_str,
        status=status,
        vendor_id=vendor_id
    )


from .. import comms
from .types import CIPMessagePreset

CIPGetHardwareRev = CIPMessagePreset(
    service=0x0e,
    class_code=0x01,
    instance=0x01,
    attribute=101
)

def get_identity(cip: comms.Driver):
    return cip.generic_message(
        service=0x01,
        class_code=0x01,
        instance=0x01,
        attribute=None
    )

def get_hardware_rev(cip: comms.Driver):
    return cip.generic_message(
        service=CIPGetHardwareRev.service,
        class_code=CIPGetHardwareRev.class_code,
        instance=CIPGetHardwareRev.instance,
        attribute=CIPGetHardwareRev.attribute,
    )

def reset(cip: comms.Driver, data):
    return cip.generic_message(
        service=0x05,
        class_code=0x01,
        instance=0x01,
        attribute=None,
        request_data=data
    )


from .. import comms
from ..common.validation import get_cip_identity

from . import types 

def get_device_info(cip: comms.Driver) -> types.CFDeviceInfo:
    cip_identity = get_cip_identity(cip)        
    return types.CFDeviceInfo(
        comms_path=cip._original_path,
        cip_identity=cip_identity,
        log=[]
    )


import os
from dataclasses import dataclass, field

from ..common.types import CIPIdentity, ResponseStatus

@dataclass
class MEArchive:
    name: str
    data: bytearray
    path: list[str]
    size: int

@dataclass
class MEFupUpgradeInfVersion:
    plat: int
    os: str
    me: str
    kep: str
    minos: str
    maxos: str
    ard: int

@dataclass
class MEFupUpgradeInfCard:
    files: list[tuple[str, str]]
    ram_size_bytes: int
    storage_size_bytes: int
    fp_size: int

@dataclass
class MEFupUpgradeInf:
    version: MEFupUpgradeInfVersion
    fwc: MEFupUpgradeInfCard
    otw: MEFupUpgradeInfCard
    drivers: list[tuple[str, int]]
    ce: list[tuple[str, str]]

@dataclass
class MEFupMEFileListInfInfo:
    me: str
    size_on_disk_bytes: int

@dataclass
class MEFupMEFileListInf:
    info: MEFupMEFileListInfInfo
    mefiles: list[str]
    
@dataclass
class MEIdentity:
    helper_version: str
    me_version: str
    major_rev: int
    minor_rev: int
    product_code: int
    product_name: str
    product_type: int
    serial_number: str = field(repr=False)
    vendor_id: int

@dataclass
class MEPaths:
    helper_file: str
    storage: str
    upload_list: str
    runtime: str
    fuwhelper_file: str

@dataclass
class MEDeviceInfo:
    comms_path: str
    cip_identity: CIPIdentity
    me_identity: MEIdentity
    log: list[str]
    files: list[str]
    running_med_file: str
    startup_mer_file: str
    me_paths: MEPaths

@dataclass
class MEResponse(object):
    device: MEDeviceInfo
    status: str



from enum import StrEnum
from warnings import warn

from .. import comms
from . import messages
from . import types 


# Known static value for successful creation of a folder.
# Further investigation needed.
CREATE_DIR_SUCCESS = 183 # 'b\xb7'

# Known functions available from RemoteHelper.
class HelperFunctions(StrEnum):
    CREATE_FILE_LIST = 'FileBrowse' # Args: {Search Path}\\*.{Search Extension}::{Results File Path}, Returns: 1 if result generated
    #CREATE_FILE_SHORTCUT = 'CreateRemShortcut' # Untested
    CREATE_FOLDER = 'CreateRemDirectory' # Args: {Folder Path}, Returns: Static value if successful [Further investigation needed]
    #CREATE_FOLDER_LIST = 'FolderBrowse' # Untested
    CREATE_ME_SHORTCUT = 'CreateRemMEStartupShortcut' # Args: {Folder Path}:{Filename.MER}: /r /delay /o /d, Returns: 1 if startup shortcut is created
    DELETE_FILE = 'DeleteRemFile' # Args: {File Path}, Returns: ???
    #DELETE_FOLDER = 'DeleteRemDirectory' # Untested
    #GET_PROCESS_RUNNING = 'IsExeRunning' # Untested
    GET_FILE_EXISTS = 'FileExists' # Args: {File Path}, Returns: 1 if {File Path} exists
    GET_FILE_SIZE = 'FileSize' # Args: {File Path}, Returns: File size in bytes
    #GET_FILE_VERSION = 'GetFileVersion' # Untested
    GET_FOLDER_EXISTS = 'StorageExists' # Args: {Folder Path}, Returns: 1 if {Folder Path} exists
    GET_FREE_SPACE = 'FreeSpace' # Args: {Folder Path}, Returns: Free size in bytes
    GET_VERSION = 'GetVersion' # Args: {File Path}, Returns: Version string
    REBOOT = 'BootTerminal' # Args: null, Returns: Comms error since upon successful execution, terminal is reset.
    #START_PROCESS = 'InvokeEXE' # Untested
    #STOP_PROCESS = 'TerminateEXE' # Untested

def run_function(cip: comms.Driver, req_args):
    """
    Executes a function on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        req_args: The request to execute.  Typically a list like ['DLL Path','Function Name','Function Args'] 

    Returns:
        resp_code (int): The message response code.
        resp_data (str): The function response data.

    Raises:
        Exception: If the message response code is not zero, indicating an error.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->N-1  | Request Arguments                                   |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Response code (typically 0 = ran, other = error)    |
        | Bytes 4->N-1  | Function response                                   |
        | Bytes N       | Null footer                                         |

    Note:
        Only a certain subset of functions are whitelisted for access
        by this method.  They are documented in the HelperFunctions enum.
    """
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.run_function(cip, req_data)
    if not resp: raise Exception(f'Failed to run function: {req_args}.')
    resp_code = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
    resp_data = resp.value[4:].decode('utf-8').strip('\x00')
    return resp_code, resp_data

def create_file_list_med(cip: comms.Driver, paths: types.MEPaths):
    req_args = [paths.helper_file,HelperFunctions.CREATE_FILE_LIST, f'\\Temp\\~MER.00\\*.med::{paths.upload_list}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True

def create_file_list_mer(cip: comms.Driver, paths: types.MEPaths):
    req_args = [paths.helper_file,HelperFunctions.CREATE_FILE_LIST, f'{paths.runtime}\\*.mer::{paths.upload_list}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True

def create_folder(cip: comms.Driver, paths: types.MEPaths, dir: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.CREATE_FOLDER, dir]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != CREATE_DIR_SUCCESS): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return True

def create_folders(cip: comms.Driver, paths: types.MEPaths, folder_path_terminal: str) -> bool:
    subfolders = folder_path_terminal.split('\\')
    current_path = subfolders[0]
    for folder in subfolders[1:]:
        current_path = f'{current_path}\\{folder}'
        if not get_folder_exists(cip=cip, paths=paths, folder_path=current_path):
            print(f'Create folder: {current_path}')
            if not create_folder(cip, paths, current_path): return False

    return True

def create_me_shortcut(cip: comms.Driver, paths: types.MEPaths, file: str, replace_comms: bool, delete_logs: bool) -> bool:
    """
    Setup a specific *.MER file to run at terminal startup.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        file (str): The *.MER file to use
        replace_comms (bool): If true, replaces the terminal comms setup with the configuration from the *.MER file.
        delete_logs (bool): If true, deletes logs at startup.

    Returns:
        Status (bool): True if function runs successfully.

    Request Format:
        {storage path}:{filename}: /r /delay /o /d

    Note:
        Even though the *.MER file resides in a subdirectory (ex: {storage path}\\Rockwell Software\\RSViewME\\Runtime\\{{filename}),
        it is only transmitted as {storage path}:{filename}.  The rest of the path is inserted by the RemoteHelper on the terminal.
    """ 
    replace_comms_param = ' /o' if replace_comms else ''
    delete_logs_param = ' /d' if delete_logs else ''
    shortcut = f'{paths.storage}:{file}: /r /delay{replace_comms_param}{delete_logs_param}'
    req_args = [paths.helper_file, HelperFunctions.CREATE_ME_SHORTCUT, shortcut]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to create ME Startup Shortcut on terminal: {file}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_file(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.DELETE_FILE, file_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete file on terminal: {file_path}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_file_list(cip: comms.Driver, paths: types.MEPaths) -> bool:
    return delete_file(cip, paths, paths.upload_list)

'''
# Untested
def delete_folder(cip: comms.Driver, paths: types.MEDevicePaths, folder_path: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.DELETE_FOLDER, folder_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return True
'''

def get_file_exists(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.GET_FILE_EXISTS, file_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_file_size(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> int:
    if not(get_file_exists(cip, paths, file_path)): raise FileNotFoundError(f'File {file_path} does not exist on remote terminal.')
    req_args = [paths.helper_file, HelperFunctions.GET_FILE_SIZE, file_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return int(resp_data)

def get_folder_exists(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.GET_FOLDER_EXISTS, folder_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_free_space(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> int:
    if not(get_folder_exists(cip, paths, folder_path)): raise FileNotFoundError(f'Folder {folder_path} does not exist on remote terminal.')
    req_args = [paths.helper_file, HelperFunctions.GET_FREE_SPACE, folder_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return int(resp_data)

def get_free_space_runtime(cip: comms.Driver, paths: types.MEPaths) -> int:
    folder_path = f'{paths.runtime}\\'
    return get_free_space(cip, paths, folder_path)

def get_version(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> str:
    if not(get_file_exists(cip, paths, file_path)): raise FileNotFoundError(f'File {file_path} does not exist on remote terminal.')
    req_args = [paths.helper_file, HelperFunctions.GET_VERSION, file_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return str(resp_data)

def reboot(cip: comms.Driver, paths: types.MEPaths):
    req_args = [paths.helper_file, HelperFunctions.REBOOT,'']
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)
    resp = messages.run_function(cip, req_data)


from collections.abc import Callable
import os
from typing import Optional
from warnings import warn

from . import comms
from .cf import dmk
from .cf import types
from .cf import validation

LOCAL_DMK_PATH = "C:\\Users\\Public\\Documents\\Rockwell Automation\\Firmware Kits"

class CFUtility(object):
    def __init__(
        self,
        comms_path: str, 
        driver: str = None, 
        ignore_terminal_valid: bool = False, 
        ignore_driver_valid: bool = False,
        local_dmk_path: str = None,
    ):
        """
        Initializes an instance of the MEUtility class.

        Args:
            comms_path (str): The path to the communications resource (ex: 192.168.1.20).
            driver (str): The driver name to use (ex: pycomm3 or pylogix).  If not specified, will default
                the first one installed that can be found.
            ignore_terminal_valid (bool): If True, ignore terminal validation checks.
            ignore_driver_valid (bool): If True, ignore driver validation checks.
            local_dmk_path (str): The default directory to assume *.DMK files are found.
        """
        self.comms_path = comms_path
        self.driver = driver
        self.ignore_terminal_valid = ignore_terminal_valid
        self.ignore_driver_valid = ignore_driver_valid
        self.local_dmk_path = LOCAL_DMK_PATH if local_dmk_path is None else local_dmk_path

    def flash_firmware(
        self, 
        dmk_path_local: str, 
        progress: Optional[Callable[[str, str, int, int], None]] = None
    ) -> types.CFResponse:
        """
        Flashes a firmware image to the remote terminal.

        Args:
            dmk_path_local (str): The local path to the firmware image file (ex: C:\\YourFolder\\\\FirmwareImage.DMK)
            progress: Optional callback for progress indication.
        """

        # Use default DMK directory if one is not specified
        if not os.path.isfile(dmk_path_local):
            if os.path.sep not in dmk_path_local:
                dmk_path_local = os.path.join(self.local_dmk_path, dmk_path_local)

        with comms.Driver(self.comms_path, self.driver) as cip:
            if (self.driver == comms.DRIVER_NAME_PYLOGIX):
                if self.ignore_driver_valid:
                    warn('Drive pylogix specified but driver validation is set to IGNORE.')
                else:
                    resp = f"""
                        Cannot flash firmware with pylogix yet, work still required on forward open/forward close.
                        Please use pycomm3 instead for now.
                    """
                    raise NotImplementedError(resp)


            # Set socket timeout first.
            # The terminal will pause at certain points and delay acknowledging messages.
            # Without this, the process will fail and the terminal will require a factory reset.
            cip.timeout = 255.0

            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_device_info(cip)

            # Perform firmware flash to terminal
            try:
                resp = dmk.process_dmk(
                    cip=cip,
                    device=self.device,
                    dmk_path_local=dmk_path_local,
                    dry_run=False,
                    progress=progress)                    
                if not(resp):
                    self.device.log.append(f'Failed to flash terminal.')
                    return types.CFResponse(self.device, types.ResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to flash terminal.')
                return types.CFResponse(self.device, types.ResponseStatus.FAILURE)

        return types.CFResponse(self.device, types.ResponseStatus.SUCCESS)


from collections.abc import Callable
import olefile
import os
from typing import Optional

from . import types

CHUNK_CONTROL_SIZE_BYTES = 2
CHUNK_DATA_SIZE_TOKENS = 16
PAGE_CONTEXT_SIZE_BYTES = 4095
PAGE_CONTROL_SIZE_BYTES = 4
PAGE_HEADER_SIZE_BYTES = 4
TOKEN_LITERAL_SIZE_BYTES = 1
TOKEN_POINTER_SIZE_BYTES = 2

STREAM_NAME_MAPPEE = '__MAPPEE'
STREAM_NAME_MAPPER = '__MAPPER'

def _get_int8_nibbles(value: int):
    high_nibble = (value >> 4) & 0x0F
    low_nibble = value & 0x0F
    return (low_nibble, high_nibble)

def _get_pointer_values(bytes: bytearray):
    (length, offset_msb) = _get_int8_nibbles(bytes[0])
    offset_lsb = bytes[1]
    offset = (offset_msb << 8) + offset_lsb
    length += 1
    return (length, offset)

def _get_expected_chunk_length(bytes) -> int:
    tokens = int.from_bytes(bytes, byteorder='little').bit_count()
    literals = CHUNK_DATA_SIZE_TOKENS - tokens
    length = (tokens * 2) + literals
    return length

def _is_pointer(control: int, index: int) -> bool:
    return bool((control >> index) & 1)

def _decompress_chunk(input: bytearray, control: int, data: bytearray, length: int) -> bytearray:
    output = bytearray(input)

    # Each chunk is comprised of a fixed number of tokens (some literal, some pointers)
    byte_index = 0
    for data_index in range(CHUNK_DATA_SIZE_TOKENS):
        if _is_pointer(control, data_index):
            token_bytes = data[byte_index:byte_index+TOKEN_POINTER_SIZE_BYTES]
            (token_length, token_offset) = _get_pointer_values(token_bytes)

            head = len(output)
            token_index = 0
            while (token_index < token_length):
                if (token_offset > 0):
                    # Normally look back and slide forward.
                    # This allows for the case where the length is
                    # greater than the offset, so part of the new bytes
                    # ends up used again within the same substitution.
                    output.append(output[head - token_offset + token_index])
                else:
                    # Offset zero is a special case where the last char
                    # is just repeated.
                    output.append(output[head - 1])
                token_index += 1

            byte_index += TOKEN_POINTER_SIZE_BYTES
        else:
            output.append(data[byte_index])
            byte_index += TOKEN_LITERAL_SIZE_BYTES
        
        if byte_index >= length: break

    return output[len(input):]

def _decompress_page(input: bytearray) -> bytearray:
    output = bytearray()
    length = len(input)
    offset = 0
    page_control_bytes = input[offset:offset + PAGE_CONTROL_SIZE_BYTES]
    offset += PAGE_CONTROL_SIZE_BYTES

    # If page is not compressed, return the rest of the page as-is
    if (page_control_bytes[0] == 0x01):
        output = input[offset:]
        return output

    # Then parse through the rest of the page
    while offset < length:
        # At the start of each chunk is a pair of control bytes
        chunk_control_bytes = input[offset:offset + CHUNK_CONTROL_SIZE_BYTES]
        chunk_control_int = int.from_bytes(chunk_control_bytes, 'little')
        if not chunk_control_bytes: break
        chunk_expected_length = _get_expected_chunk_length(chunk_control_bytes)

        offset += CHUNK_CONTROL_SIZE_BYTES
        chunk_data_bytes = input[offset:offset + chunk_expected_length]
        chunk_actual_length = len(chunk_data_bytes)
        offset += chunk_expected_length

        context = output[-PAGE_CONTEXT_SIZE_BYTES:]
        output += _decompress_chunk(context, chunk_control_int, chunk_data_bytes, chunk_actual_length)

    # Return decompressed page bytes
    return output

def _decompress_stream(
    input: bytearray,
    progress_desc: str = None,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> bytearray:
    output = bytearray()
    length = len(input)
    offset = 0
    while offset < length:
        # At the start of each page there is an 4 byte header to signify page length
        page_size = int.from_bytes(input[offset:offset + PAGE_HEADER_SIZE_BYTES], byteorder='little')
        offset += PAGE_HEADER_SIZE_BYTES
        page_bytes = input[offset:offset + page_size]
        offset += page_size

        output += _decompress_page(page_bytes)
        if progress:
            desc = f'Decompressing'
            if progress_desc: desc += f' {progress_desc}'
            progress(desc, 'bytes', length, offset)

    return output

def _create_subfolders(output_path: str, archive_paths: list[str]):
    folders = archive_paths[:-1]
    current_path = output_path
    for folder in folders:
        current_path = os.path.join(current_path, folder)
        os.makedirs(current_path, exist_ok=True)
    return os.path.join(current_path, archive_paths[-1])

def _get_mapper_filename(input: bytearray) -> str:
    offset = 0
    length = int.from_bytes(input[offset:offset + PAGE_HEADER_SIZE_BYTES], byteorder='little')
    offset += PAGE_HEADER_SIZE_BYTES
    name = input[offset:].decode('utf-16-le').rstrip('\x00')
    return name

def _get_mapper_for_mappee(ole: olefile.OleFileIO, mappee_name: str) -> str:
    # The assumption is that if there are multiple MAPPER/MAPPEE pairs
    # in the archive, that they have unique numbers at end of the stream name.
    mapper_name = mappee_name.replace(STREAM_NAME_MAPPEE, STREAM_NAME_MAPPER)
    mapper_data = ole.openstream(mapper_name).read()
    return _get_mapper_filename(mapper_data)

def decompress_archive(
    ole: olefile.OleFileIO,
    progress: Optional[Callable[[str, str, int, int], None]] = None
) -> list[types.MEArchive]:
    streams = []
    for stream_path in ole.listdir():
        stream_name = '/'.join(stream_path)
        if (ole.exists(stream_name) and not ole.get_type(stream_name) == olefile.STGTY_STORAGE):
            original_name = stream_name
            # If a stream name starts with __MAPPEE it has the content of the file.
            # If a stream name starts with __MAPPER it has the name of the file.
            #
            # This logic restores the name from the MAPPER to the MAPPEE and
            # excludes the MAPPER from the stream list.
            if stream_name.startswith(STREAM_NAME_MAPPEE):
                actual_name = _get_mapper_for_mappee(ole, original_name)
                stream_name = actual_name
                stream_path[-1] = actual_name
            if stream_name.startswith(STREAM_NAME_MAPPER):
                continue
            
            stream_data = ole.openstream(original_name).read()
            try:
                stream_data = _decompress_stream(
                    input=stream_data,
                    progress_desc=stream_name,
                    progress=progress
                )
            except Exception as e:
                # Some streams aren't compressed.
                #
                # Is there a better way to retain them and still
                # print exceptions for failed decompressions?
                #print(e)
                pass

            stream_info = types.MEArchive(
                name=stream_name,
                data=stream_data,
                path=stream_path,
                size=len(stream_data)
            )
                
            streams.append(stream_info)
    return streams

def archive_to_disk(file_path: str, output_path: str):
    # Directly dump archive with no application-specific
    # handling (ex: for *.FUP or *.MER)

    # Create output folder if it doesn't exist yet
    if not(os.path.exists(output_path)): os.makedirs(output_path, exist_ok=True)

    with olefile.OleFileIO(file_path) as ole:
        streams = decompress_archive(ole)
        for stream in streams:
            stream_output_path = _create_subfolders(output_path, stream.path)
            with open(stream_output_path, 'wb') as f:
                f.write(stream.data)


from warnings import warn

# see which drivers are installed on the system
AVAILABLE_DRIVERS = []
DRIVER_NAME_PYCOMM3 = 'pycomm3'
DRIVER_NAME_PYLOGIX = 'pylogix'
try:
    from pycomm3 import CIPDriver, const, util
    AVAILABLE_DRIVERS.append(DRIVER_NAME_PYCOMM3)
except: pass

try:
    from pylogix import PLC
    AVAILABLE_DRIVERS.append(DRIVER_NAME_PYLOGIX)
except: pass

class Driver:

    def __init__(self, comms_path=None, driver=None):
        self._original_path = comms_path

        if not AVAILABLE_DRIVERS:
            raise ImportError("You need to install pycomm3 or pylogix")

        # select the driver the user requested
        if driver:
            if driver in AVAILABLE_DRIVERS:
                self._driver = driver
            else:
                raise ImportError(f"{driver} is not installed on the system")
        else:
            # no driver requested, pick the first one
            self._driver = AVAILABLE_DRIVERS[0]

        # Split originally supplied path into IP address and route if needed
        if is_routed_path(self._original_path):
            self._ip_address, self._route_path = convert_path_pycomm3_to_pylogix(self._original_path)
        else:
            self._ip_address = self._original_path
            self._route_path = None

        # Configure driver
        if self._driver == DRIVER_NAME_PYLOGIX:
            self.cip = PLC(self._ip_address)
            self.cip.Route = self._route_path
        elif self._driver == DRIVER_NAME_PYCOMM3:
            self.cip = CIPDriver(self._original_path)
            self._const_timeout_ticks = const.TIMEOUT_TICKS # Used to check for unconnected send firmware upgrade
            self.cip.open()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._driver == DRIVER_NAME_PYLOGIX:
            self.cip.Close()
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip.close()

    def generic_message(self, service, class_code, instance, attribute, request_data=b'', connected=False):
        if self._driver == DRIVER_NAME_PYLOGIX:
            ret = self.cip.Message(cip_service=service,
                                   cip_class=class_code,
                                   cip_instance=instance,
                                   cip_attribute=attribute,
                                   data=request_data)
            if ret.Status == "Success":
                status = None
            else:
                status = ret.Status
            return Response(ret.Value[44:], None, status)
        elif self._driver == DRIVER_NAME_PYCOMM3:
            if is_routed_path(self._original_path):
                unconnected_send = True
                route_path = True
            else:
                unconnected_send = False
                route_path = False
            return self.cip.generic_message(service=service,
                                            class_code=class_code,
                                            instance=instance,
                                            attribute=attribute,
                                            request_data=request_data,
                                            connected=connected,
                                            unconnected_send=unconnected_send,
                                            route_path=route_path
                                            )

    @property
    def connection_size(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            raise self.cip.connection_size
        if self._driver == DRIVER_NAME_PYLOGIX:
            return self.cip.ConnectionSize

    @connection_size.setter
    def connection_size(self, new_value):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip._cfg['connection_size'] = new_value
            self.cip.close()
            self.cip.open()
        if self._driver == DRIVER_NAME_PYLOGIX:
            self.cip.ConnectionSize = new_value

    @property
    def timeout(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            return self.cip._cfg['socket_timeout']
        if self._driver == DRIVER_NAME_PYLOGIX:
            return self.cip.SocketTimeout

    @timeout.setter
    def timeout(self, new_value):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip._cfg['timeout'] = new_value
            self.cip._cfg['socket_timeout'] = new_value
            self.cip.close()
            self.cip.open()
        if self._driver == DRIVER_NAME_PYLOGIX:
            self.cip.SocketTimeout = new_value

    def open(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip.open()

    def close(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip.close()

    def sequence_reset(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip._sequence = util.cycle(65535, start=1)
        if self._driver == DRIVER_NAME_PYLOGIX:
            self.cip.conn._sequence_counter = 1
        
    def forward_open(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip._forward_open()
        if self._driver == DRIVER_NAME_PYLOGIX:
            return
            self.cip.conn._connect(True)

    def forward_close(self):
        if self._driver == DRIVER_NAME_PYCOMM3:
            self.cip._forward_close()
        if self._driver == DRIVER_NAME_PYLOGIX:
            return
            self.cip.conn._close_connection()

    @property
    def me_chunk_size(self):
        return get_me_chunk_size(self._original_path)

def get_me_chunk_size(path: str) -> int:
    # When files are transferred using ME services, this is the maximum
    # number of bytes used per one message.
    if is_routed_path(path):
        # Routed path through one or more other devices (ex: through CLX rack and then to terminal)
        ip_address, route_path = convert_path_pycomm3_to_pylogix(path)

        max_size = 466
        working_size = max_size - 2 # Remove route path size and reserved bytes
        for segment in route_path:
            port = segment[0]
            try:
                path = int(segment[1])
                path_size = 1
                segment_size = 1 # 1 control byte
            except:
                path = str(segment[1])
                path_size = len(path)
                segment_size = 2 + path_size # 1 control byte, 1 length byte, X path bytes
                if segment_size % 2: segment_size += 1 # if segment length is odd, there is a pad byte

            working_size -= segment_size
        return working_size
    else:
        # Direct path
        # Tests up to 2000 bytes did succeed, >2000 bytes failed.
        return 1984

def is_routed_path(path: str) -> bool:
    if (',' in path) or ('/' in path) or ('\\' in path):
        return True
    else:
        return False
        
def convert_path_pycomm3_to_pylogix(path: str):
    """
    Converts a pycomm3-style route string into a pylogix Route list.

    Args:
        path (str): A route string, e.g., '192.168.2.10/bp/3/enet/192.168.1.20'

    Returns:
        tuple: (starting_ip: str, pylogix_route: list of tuples)
    """
    # make sure we are only working with commas, then split
    parts = path.replace("/", ",").replace("\\", ",").split(",")
    ip_address = parts.pop(0)
    route = []

    if parts:
        # make sure even number of segments
        if len(parts) % 2:
            raise ValueError("Path must have at least one routing pair (port, destination) after the start IP.")
        
        for i in range(len(parts)):
            # try to convert each path segment to an int
            try:
                parts[i] = int(parts[i])
            except:
                if parts[i] == "backplane":
                    parts[i] = 1
                elif parts[i] == "bp":
                    parts[i] = 1
                elif parts[i] == "enet":
                    parts[i] = 2

        # convert the route to pylogix format (2 item lists)
        route = [tuple(parts[i:i+2]) for i in range(0, len(parts), 2)]
    return ip_address, route

class Response(object):
    def __init__(self, value, type, error):
        self.tag = 'generic'
        self.value = value
        self.type = type
        self.error = error


from .. import comms
from ..common.types import CIPMessagePreset

MECreateTransfer = CIPMessagePreset(
    service=int.from_bytes(b'\x08', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=0x00,
    attribute=None
)

MEDeleteTransfer = CIPMessagePreset(
    service=int.from_bytes(b'\x09', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=None,
    attribute=None
)

MEReadFileChunk = CIPMessagePreset(
    service=int.from_bytes(b'\x53', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=None,
    attribute=None
)

MEReadFileReady = CIPMessagePreset(
    service=int.from_bytes(b'\x0e', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=0x00,
    attribute=None
)

MEReadRegistry = CIPMessagePreset(
    service=int.from_bytes(b'\x51', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFE', byteorder='big'),
    instance=0x00,
    attribute=None
)

MERunFunction = CIPMessagePreset(
    service=int.from_bytes(b'\x50', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFD', byteorder='big'),
    instance=0x00,
    attribute=None
)

MEWriteFileChunk = CIPMessagePreset(
    service=int.from_bytes(b'\x52', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=None,
    attribute=None
)

MEWriteFileReady = CIPMessagePreset(
    service=int.from_bytes(b'\x10', byteorder='big'),
    class_code=int.from_bytes(b'\x04\xFF', byteorder='big'),
    instance=0x01,
    attribute=None
)

def create_transfer(cip: comms.Driver, data):
    return cip.generic_message(
        service=MECreateTransfer.service,
        class_code=MECreateTransfer.class_code,
        instance=MECreateTransfer.instance,
        attribute=MECreateTransfer.attribute,
        request_data=data
    )

def delete_transfer(cip: comms.Driver, instance: int):
    return cip.generic_message(
        service=MEDeleteTransfer.service,
        class_code=MEDeleteTransfer.class_code,
        instance=instance,
        attribute=MEDeleteTransfer.attribute
    )

def read_file_chunk(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=MEReadFileChunk.service,
        class_code=MEReadFileChunk.class_code,
        instance=instance,
        attribute=MEReadFileChunk.attribute,
        request_data=data
    )

def read_registry(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEReadRegistry.service,
        class_code=MEReadRegistry.class_code,
        instance=MEReadRegistry.instance,
        attribute=MEReadRegistry.attribute,
        request_data=data
    )

def run_function(cip: comms.Driver, data):
    return cip.generic_message(
        service=MERunFunction.service,
        class_code=MERunFunction.class_code,
        instance=MERunFunction.instance,
        attribute=MERunFunction.attribute,
        request_data=data
    )

def write_file_chunk(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=MEWriteFileChunk.service,
        class_code=MEWriteFileChunk.class_code,
        instance=instance,
        attribute=MEWriteFileChunk.instance,
        request_data=data
    )

def read_file_ready(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEReadFileReady.service,
        class_code=MEReadFileReady.class_code,
        instance=MEReadFileReady.instance,
        attribute=MEReadFileReady.attribute,
        request_data=data
    )

def write_file_ready(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEWriteFileReady.service,
        class_code=MEWriteFileReady.class_code,
        instance=MEWriteFileReady.instance,
        attribute=MEWriteFileReady.attribute,
        request_data=data
    )